# Copyright (c) 2015, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.
"""
The signals module contains code which supports the registration of signal
handlers, as well as the emitting of signals.

Attribution:
    Much of the code here was inspired by Django signals:
    https://github.com/django/django/blob/1.8.3/django/dispatch/dispatcher.py
"""
import collections
import logging
import threading
import weakref

from .vendor import six

try:
    from weakref import WeakMethod
except ImportError:
    from weakrefmethod import WeakMethod


# Module-level logger
LOG = logging.getLogger(__name__)

# Signal receiver functions
__receivers = collections.defaultdict(list)

# Synchronize access to __receivers
__lock = threading.Lock()


def __is_dead(ref):
    """Return ``True`` if the weak reference `ref` no longer points to an
    in-memory object.
    """
    return ref() is None


def __make_id(receiver):
    """Generate an identifier for a callable signal receiver.

    This is used when disconnecting receivers, where we need to correctly
    establish equivalence between the input receiver and the receivers assigned
    to a signal.

    Args:
        receiver: A callable object.

    Returns:
        An identifier for the receiver.
    """
    if __is_bound_method(receiver):
        return (id(receiver.__func__), id(receiver.__self__))
    return id(receiver)


def __purge():
    """Remove all dead signal receivers from the global receivers collection.

    Note:
        It is assumed that the caller holds the __lock.
    """
    global __receivers
    newreceivers = collections.defaultdict(list)

    for signal, receivers in six.iteritems(__receivers):
        alive = [x for x in receivers if not __is_dead(x)]
        newreceivers[signal] = alive

    __receivers = newreceivers


def __live_receivers(signal):
    """Return all signal handlers that are currently still alive for the
    input `signal`.

    Args:
        signal: A signal name.

    Returns:
        A list of callable receivers for the input signal.
    """
    with __lock:
        __purge()
        receivers = [funcref() for funcref in __receivers[signal]]

    return receivers


def __is_bound_method(method):
    """Return ``True`` if the `method` is a bound method (attached to an class
    instance.

    Args:
        method: A method or function type object.
    """
    if not(hasattr(method, "__func__") and hasattr(method, "__self__")):
        return False

    # Bound methods have a __self__ attribute pointing to the owner instance
    return six.get_method_self(method) is not None


def __check_receiver(receiver):
    """Check that the `receiver` is an acceptable signal receiver.

    Signal receivers must be one of the following:

    * Callable objects.
    * Bound instance methods
    * Static/class methods
    * Functions

    Raises:
        TypeError: If `func` is not an appropriate signal receiver.
    """
    if six.callable(receiver):
        return

    error = ("Signal receivers must be functions, callable objects, or "
             "static/class/bound methods.")
    raise TypeError(error)


def connect(signal, receiver):
    """Register `receiver` method/function as a receiver for the `signal`.

    When the signal is emitted, this receiver will be invoked along with
    all other associated signals.

    Args:
        signal: A signal identifier (e.g., a signal name)
        receiver: A callable object to connect to the signal.
    """
    __check_receiver(receiver)

    if __is_bound_method(receiver):
        ref = WeakMethod
    else:
        ref = weakref.ref

    with __lock:
        __purge()
        __receivers[signal].append(ref(receiver))


def disconnect(signal, receiver):
    """Disconnect the receiver `func` from the signal, identified by
    `signal_id`.

    Args:
        signal: The signal identifier.
        receiver: The callable receiver to disconnect.

    Returns:
        True if the receiver was successfully disconnected. False otherwise.
    """
    inputkey = __make_id(receiver)

    with __lock:
        __purge()
        receivers = __receivers.get(signal)

        for idx in six.moves.range(len(receivers)):
            connected = receivers[idx]()

            if inputkey != __make_id(connected):
                continue

            del receivers[idx]
            return True  # receiver successfully disconnected!

    return False


def receiver(signal):
    """Function decorator which registers the wrapped function as a signal
    receiver for the signal `signal_name`.

    Warning:
        This will not work with unbound instance methods.

    Args:
        signal: A signal identifier (e.g., a name).
    """
    def decorator(func):
        connect(signal, func)  # register the signal receiver
        return func
    return decorator


def emit(signal, *args, **kwargs):
    """Emit a signal by serially calling each registered signal receiver for
    the `signal`.

    Note:
        The receiver must accept the *args and/or **kwargs that have been
        passed to it. There expected parameters are not dictated by
        mixbox.
        
    Args:
        signal: A signal identifier or name.
        *args: A variable-length argument list to pass to the receiver.
        **kwargs: Keyword-arguments to pass to the receiver.
    """
    if signal not in __receivers:
        return

    receivers = __live_receivers(signal)

    for func in receivers:
        func(*args, **kwargs)



# Copyright (c) 2015, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

"""
Common routines for working with Python objects.
"""
import importlib

from .vendor import six


def is_sequence(value):
    """Determine if a value is a sequence type.

    Returns:
      ``True`` if `value` is a sequence type (e.g., ``list``, or ``tuple``).
      String types will return ``False``.

    NOTE: On Python 3, strings have the __iter__ defined, so a simple hasattr
    check is insufficient.
    """
    return (hasattr(value, "__iter__") and not
            isinstance(value, (six.string_types, six.binary_type)))


def import_class(classpath):
    """Import the class referred to by the fully qualified class path.

    Args:
        classpath: A full "foo.bar.MyClass" path to a class definition.

    Returns:
        The class referred to by the classpath.

    Raises:
        ImportError: If an error occurs while importing the module.
        AttributeError: IF the class does not exist in the imported module.
    """
    modname, classname = classpath.rsplit(".", 1)
    module = importlib.import_module(modname)
    klass  = getattr(module, classname)
    return klass


def resolve_class(classref):
    """Attempt to return a Python class for the input class reference.

    If `classref` is a class or None, return it. If `classref` is a
    python classpath (e.g., "foo.bar.MyClass") import the class and return
    it.

    Args:
        classref: A fully-qualified Python path to class, or a Python class.

    Returns:
        A class.
    """
    if classref is None:
        return None
    elif isinstance(classref, six.class_types):
        return classref
    elif isinstance(classref, six.string_types):
        return import_class(classref)
    else:
        raise ValueError("Unable to resolve class for '%s'" % classref)


class classproperty(object):
    """A ``property`` descriptor that works on classes rather than
    instances.

    Source:
        http://stackoverflow.com/a/3203659
    """

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


def needkwargs(*argnames):
    """Function decorator which checks that the decorated function is called
    with a set of required kwargs.

    Args:
        *argnames: String keyword argument names.

    Raises:
        ValueError: If a required kwarg is missing in the decorated function
            call.
    """
    required = set(argnames)

    def decorator(func):
        def inner(*args, **kwargs):
            missing = required - set(kwargs)
            if missing:
                err = "%s kwargs are missing." % list(missing)
                raise ValueError(err)
            return func(*args, **kwargs)
        return inner
    return decorator



# Copyright (c) 2015, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.
"""
Entity field data descriptors (TypedFields) and associated classes.
"""
import functools
import inspect

from .datautils import is_sequence, resolve_class
from .typedlist import TypedList
from .dates import parse_date, parse_datetime, serialize_date, serialize_datetime
from .xml import strip_cdata, cdata
from .vendor import six
from .compat import long


def unset(entity, *types):
    """Unset the TypedFields on the input `entity`.

    Args:
        entity: A mixbox.Entity object.
        *types: A variable-length list of TypedField subclasses. If not
            provided, defaults to TypedField.
    """
    if not types:
        types = (TypedField,)

    fields = list(entity._fields.keys())
    remove = (x for x in fields if isinstance(x, types))

    for field in remove:
        del entity._fields[field]


def _matches(field, params):
    """Return True if the input TypedField `field` contains instance attributes
    that match the input parameters.

    Args:
        field: A TypedField instance.
        params: A dictionary of TypedField instance attribute-to-value mappings.

    Returns:
        True if the input TypedField matches the input parameters.
    """
    fieldattrs = six.iteritems(params)
    return all(getattr(field, attr) == val for attr, val in fieldattrs)


def iterfields(klass):
    """Iterate over the input class members and yield its TypedFields.

    Args:
        klass: A class (usually an Entity subclass).

    Yields:
        (class attribute name, TypedField instance) tuples.
    """
    is_field = lambda x: isinstance(x, TypedField)

    for name, field in inspect.getmembers(klass, predicate=is_field):
        yield name, field


def find(entity, **kwargs):
    """Return all TypedFields found on the input `Entity` that were initialized
    with the input **kwargs.

    Example:
        >>> find(myentity, multiple=True, type_=Foo)

    Note:
        TypedFields.__init__() can accept a string or a class as a type_
        argument, but this method expects a class.

    Args:
        **kwargs: TypedField __init__ **kwargs to search on.

    Returns:
        A list of TypedFields with matching **kwarg values.
    """
    try:
        typedfields = entity.typed_fields()
    except AttributeError:
        typedfields = iterfields(entity.__class__)

    matching = [x for x in typedfields if _matches(x, kwargs)]
    return matching


class TypedField(object):

    def __init__(self, name, type_=None,
                 key_name=None, comparable=True, multiple=False,
                 preset_hook=None, postset_hook=None, factory=None,
                 listfunc=None):
        """
        Create a new field.

        Args:
            `name` (str): name of the field as contained in the binding class.
            `type_` (type/str): Required type for values assigned to this field.
                If`None`, no type checking is performed. String values are
                treated as fully qualified package paths to a class (e.g.,
                "A.B.C" would be the full path to the type "C".)
            `key_name` (str): name for field when represented as a dictionary.
                (Optional) If omitted, `name.lower()` will be used.
            `comparable` (boolean): whether this field should be considered
                when checking Entities for equality. Default is True. If False,
                this field is not considered.
            `multiple` (boolean): Whether multiple instances of this field can
                exist on the Entity.
            `preset_hook` (callable): called before assigning a value to this
                field, but after type checking is performed (if applicable).
                This should typically be used to perform additional validation
                checks on the value, perhaps based on current state of the
                instance. The callable should accept two arguments: (1) the
                instance object being modified, and (2)the value it is being
                set to.
            `postset_hook` (callable): similar to `preset_hook` (and takes the
                same arguments), but is called after setting the value. This
                can be used, for example, to modify other fields of the
                instance to maintain some type of invariant.
            `listfunc` (callable): A datatype or a function that creates a
                mutable sequence type for multiple field internal storage.
                E.g., "list".
        """
        self.name = name
        self.comparable = comparable
        self.multiple = multiple
        self.preset_hook = preset_hook
        self.postset_hook = postset_hook

        # The type of the field. This is lazily set via the type_ property
        # at first access.
        self._unresolved_type = type_

        # The factory for the field. This controls which class will be used
        # for from_dict() and from_obj() calls for this field.
        # Lazily set via the factory property.
        self._unresolved_factory = factory

        # Dictionary key name for the field.
        if key_name:
            self._key_name = key_name
        else:
            self._key_name = name.lower()

        # List creation function for multiple fields.
        if listfunc:
            self._listfunc = listfunc
        elif type_:
            self._listfunc = functools.partial(TypedList, type=type_)
        else:
            self._listfunc = list

    def __get__(self, instance, owner=None):
        """Return the TypedField value for the input `instance` and `owner`.

        If the TypedField is a "multiple" field and hasn't been set yet,
        set the field to an empty list and return it.

        Args:
            instance: An instance of the `owner` class that this TypedField
                belongs to..
            owner: The TypedField owner class.
        """
        if instance is None:
            return self
        elif self in instance._fields:
            return instance._fields[self]
        elif self.multiple:
            return instance._fields.setdefault(self, self._listfunc())
        else:
            return None

    def _clean(self, value):
        """Validate and clean a candidate value for this field."""
        if value is None:
            return None
        elif self.type_ is None:
            return value
        elif self.check_type(value):
            return value
        elif self.is_type_castable:  # noqa
            return self.type_(value)

        error_fmt = "%s must be a %s, not a %s"
        error = error_fmt % (self.name, self.type_, type(value))
        raise TypeError(error)

    def __set__(self, instance, value):
        """Sets the field value on `instance` for this TypedField.

        If the TypedField has a `type_` and `value` is not an instance of
        ``type_``, an attempt may be made to convert `value` into an instance
        of ``type_``.

        If the field is ``multiple``, an attempt is made to convert `value`
        into a list if it is not an iterable type.
        """
        if self.multiple:
            if value is None:
                value = self._listfunc()
            elif not is_sequence(value):
                value = self._listfunc([self._clean(value)])
            else:
                value = self._listfunc(self._clean(x) for x in value if x is not None)
        else:
            value = self._clean(value)

        if self.preset_hook:
            self.preset_hook(instance, value)

        instance._fields[self] = value

        if self.postset_hook:
            self.postset_hook(instance, value)

    def __str__(self):
        return self.name

    def check_type(self, value):
        if not self.type_:
            return True
        elif hasattr(self.type_, "istypeof"):
            return self.type_.istypeof(value)
        else:
            return isinstance(value, self.type_)

    @property
    def key_name(self):
        return self._key_name

    @property
    def type_(self):
        try:
            return self._resolved_type
        except AttributeError:
            self._resolved_type = resolve_class(self._unresolved_type)
        return self._resolved_type

    @type_.setter
    def type_(self, value):
        self._resolved_type = value

    @property
    def factory(self):
        try:
            return self._resolved_factory
        except AttributeError:
            self._resolved_factory = resolve_class(self._unresolved_factory)
        return self._resolved_factory

    @factory.setter
    def factory(self, value):
        self._resolved_factory = value

    @property
    def transformer(self):
        """Return the class for this field that transforms non-Entity objects
        (e.g., dicts or binding objects) into Entity instances.

        Any non-None value returned from this method should implement a
        from_obj() and from_dict() method.

        Returns:
            None if no type_ or factory is defined by the field. Return a class
            with from_dict and from_obj methods otherwise.
        """
        if self.factory:
            return self.factory
        elif self.type_:
            return self.type_
        else:
            return None

    @property
    def is_type_castable(self):
        return getattr(self.type_, "_try_cast", False)

    def binding_value(self, value):
        return value

    def dict_value(self, value):
        return value

    def __copy__(self):
        """See __deepcopy__."""
        return self

    def __deepcopy__(self, memo):
        """Return itself (don't actually make a copy at all).

        TypedFields store themselves as a key in an Entity._fields dictionary
        and use themselves as a key for value retrieval.

        The deepcopy() function would normally descend into the _fields dictionary
        of an Entity and replace the keys with *copies* of the original
        TypedFields.

        As such, a TypedField would never find itself in a deepcopied Entity,
        because the _fields dictionary had its keys swapped out for copies
        of the original TypedField.

        We could control __deepcopy__ at the Entity level, but it's a fair
        amount more complicated and ultimately, we probably never want
        TypedFields to actually be copied since they are class-level
        property descriptors.
        """
        memo[id(self)] = self  # add self to the memo so this isn't called again.
        return self


class BytesField(TypedField):
    def _clean(self, value):
        return six.binary_type(value)


class TextField(TypedField):
    def _clean(self, value):
        return six.text_type(value)


class BooleanField(TypedField):
    def _clean(self, value):
        return bool(value)


class IntegerField(TypedField):
    def _clean(self, value):
        if value in (None, ""):
            return None
        elif isinstance(value, six.string_types):
            return int(value, 0)
        else:
            return int(value)


class LongField(TypedField):
    def _clean(self, value):
        if value in (None, ""):
            return None
        elif isinstance(value, six.string_types):
            return long(value, 0)
        else:
            return long(value)


class FloatField(TypedField):
    def _clean(self, value):
        if value in (None, ""):
            return None
        return float(value)


class DateTimeField(TypedField):
    def _clean(self, value):
        return parse_datetime(value)

    def dict_value(self, value):
        return serialize_datetime(value)

    def binding_value(self, value):
        return serialize_datetime(value)


class DateField(TypedField):
    def _clean(self, value):
        return parse_date(value)

    def dict_value(self, value):
        return serialize_date(value)

    def binding_value(self, value):
        return serialize_date(value)


class CDATAField(TypedField):
    def _clean(self, value):
        return strip_cdata(value)

    def binding_value(self, value):
        return cdata(value)


class IdField(TypedField):
    def __set__(self, instance, value):
        """Set the id field to `value`. If `value` is not None or an empty
        string, unset the idref fields on `instance`.
        """
        super(IdField, self).__set__(instance, value)

        if value:
            unset(instance, IdrefField)


class IdrefField(TypedField):
    def __set__(self, instance, value):
        """Set the idref field to `value`. If `value` is not None or an empty
        string, unset the id fields on `instance`.
        """
        super(IdrefField, self).__set__(instance, value)

        if value:
            unset(instance, IdField)
